// Compile with :
// g++ -g main.cc -o main -ldl -lpthread
// g++ -g shared_object.cc -fPIC -shared -o shared_object.so -lpthread

// Make it crash with : NO_KEY_DELETE=1 ./main
// Make it work with : ./main

#include <assert.h>
#include <cstdlib>
#include <dlfcn.h>
#include <pthread.h>
#include <stdio.h>
#include <unistd.h>

static void * plugin_handle = NULL;

void unload_plugin() {
    printf(">unload_plugin\n");
    assert(plugin_handle);
    assert(dlclose(plugin_handle) == 0);
    printf("<unload_plugin\n");
}

static void (*callable)(void) = NULL;

static int go_on = 1;
static int work_done = 0;

static void * thread_start(void *arg)
{
    callable();
    work_done = 1;
    
    /* simulate a worker thread that would wait for more orders */
    while(go_on)
        usleep(5 * 1000);
    return NULL;
}

static     pthread_t thread_id = 0;
 
static void   __attribute__ ((destructor)) finalizer(void)  
{
    printf(">main_finalizer\n");
    void* res;
    /* ask the worker thread to terminate */
    go_on = 0;
    pthread_join(thread_id, &res);
    printf("<main_finalizer\n");
}

int main(int argc, char* argv[])
{
    plugin_handle = dlopen("./shared_object.so",RTLD_LAZY);
    if (plugin_handle)
        std::atexit(&unload_plugin);
    callable = (void(*)(void))dlsym(plugin_handle, "do_work");
    
    pthread_attr_t attr;
    assert(pthread_attr_init(&attr) == 0);
    pthread_create(&thread_id, &attr, &thread_start, NULL);
    while(!work_done)
        usleep(5 * 1000);
    
    return 0;
}
